package it.polito.bigdata.spark.example;

import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.sql.types.DataTypes;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;


public class SparkDriver {
	
	public static void main(String[] args) {

		// The following two lines are used to switch off some verbose log messages
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);


		String housePath, consumptionPath;
		String outputPath1, outputPath2;

		housePath = args[0];
		consumptionPath = args[1];
		outputPath1 = args[2];
		outputPath2 = args[3];



		// Create a Spark Session object and set the name of the application
		SparkSession ss = SparkSession.builder().appName("Exam20230215 Spark")
									.master("local").getOrCreate();

		// Define the dataframes associated with the input files
		Dataset<Row> houseDF = ss.read().format("csv")
				.option("header", false).option("inferSchema", true)
				.load(housePath)
				.withColumnRenamed("_c0", "HID")
				.withColumnRenamed("_c1", "City")
				.withColumnRenamed("_c2", "Country")
				.withColumnRenamed("_c3", "YearBuilt");


		// Register the temporary table houses 
		houseDF.createOrReplaceTempView("houses");
	
		Dataset<Row> consumptionDF = ss.read().format("csv")
				.option("header", false).option("inferSchema", true)
				.load(consumptionPath)
				.withColumnRenamed("_c0", "HID")
				.withColumnRenamed("_c1", "Month")
				.withColumnRenamed("_c2", "M3");

		// Register the temporary table houses 
		consumptionDF.createOrReplaceTempView("consumption");


		
		// Part 1

		// Register a function that given a month returns the trimester
		ss.udf().register("TRIMESTER", 
				(String month) -> (Integer.parseInt(month.split("/")[1])-1)/3,
				DataTypes.IntegerType);		

		// Register a function that given a month returns the year
		ss.udf().register("EXTRACTYEAR", 
				(String month) -> Integer.parseInt(month.split("/")[0]),
				DataTypes.IntegerType);		


		// Select the water consumptions only for years 2022 and 2021
		// Computere the water consumption for each trimester of the years 2022 and 2021
		Dataset<Row> waterConsPerTrimDF = 
				ss.sql("SELECT HID, TRIMESTER(Month) AS Trim, EXTRACTYEAR(Month) AS Year, SUM(M3) AS TrimCons "
			  		+ "FROM consumption "
					+ "WHERE EXTRACTYEAR(Month)=2021 OR EXTRACTYEAR(Month)=2022 "
					+ "GROUP BY HID, TRIMESTER(Month), EXTRACTYEAR(Month)");

		// Register the temporary table waterConsPerTrim 
		waterConsPerTrimDF.createOrReplaceTempView("waterConsPerTrim");

		// Register a function that given a year returns -1 if ther year is 2021. +1 otherwise.
		ss.udf().register("CONSYEAR", 
			(Integer year) -> {if (year==2021) 
									return -1;
								else 
								 	return +1;},
			DataTypes.IntegerType);	

		// Select the trimesters for each house with a consumption in the year 2022
		// greater than the consumption in the year 2021
		Dataset<Row> housesComparisonTrimConsDF = 
				ss.sql("SELECT HID, Trim, SUM(CONSYEAR(Year)*TrimCons) "
					+ "FROM waterConsPerTrim "
					+ "GROUP BY HID, Trim "
					+ "HAVING SUM(CONSYEAR(Year)*TrimCons)>0 ");


		// Register the temporary table housesComparisonTrimCons 
		housesComparisonTrimConsDF.createOrReplaceTempView("housesComparisonTrimCons");

		// Select the houses, and theirs cities, if they have an increasing consumption
		// in at least 3 trimesters
 		Dataset<Row> housesWithIncreasingConsDF = 
				ss.sql("SELECT Houses.HID, City "
					+ "FROM housesComparisonTrimCons, Houses "
					+ "WHERE housesComparisonTrimCons.HID=Houses.HID "
					+ "GROUP BY Houses.HID, City "
					+ "HAVING COUNT(*)>=3"); 

		housesWithIncreasingConsDF.write()
			.format("csv")
			.option("header", false)
			.save(outputPath1);



		// Part 2

		// Compute the annual consumption for each house in each year
		Dataset<Row> waterConsPerYearAndHouseDF = 
				ss.sql("SELECT HID, EXTRACTYEAR(Month) AS Year, SUM(M3) AS AnnualCons "
			  		+ "FROM consumption "
					+ "GROUP BY HID, EXTRACTYEAR(Month)");

		// Register the temporary table waterConsPerYearAndHouse 
		waterConsPerYearAndHouseDF.createOrReplaceTempView("waterConsPerYearAndHouse");


		// Define the "elements of the windows" of two consecutive years for each house
		// Given a tuple HID, Year, AnnualConsumption 
		// the query returns two tuples:
		// - HID,   Year,  AnnualConsumption  <- Second element/tuple of the window ending at year "Year" for house HID 
		// - HID, Year+1, -AnnualConsumption  <- First element/tuple of the window ending at year "Year+1" for house HID 
		Dataset<Row> elementsWindowsDF = 
				ss.sql("SELECT HID, Year, AnnualCons "
			  		+ "FROM waterConsPerYearAndHouse "
					+ "UNION ALL "
					+ "SELECT HID, Year+1 AS Year, -1*AnnualCons AS AnnualCons "
					+ "FROM waterConsPerYearAndHouse ");

		// Register the temporary table elementsWindows 
		elementsWindowsDF.createOrReplaceTempView("elementsWindows");


		// Select the pairs (HID,Year) associated with windows with a decreasing consumption 
		// (if the sum of the consumptions in a window is negative it means that the consumption
		//  in the first year of the window is greater than the consumption in the second year 
		//  of the window). Consider only complete windows (i.e., windows with 2 elements).
		// 
		// Each HID can be associated with many windows with a decreasing trend. 
		// Distint HIDs must be selected to compute the correct result. 
		Dataset<Row> housesWithAtLeastOneDecreaseDF = 
				ss.sql("SELECT DISTINCT HID "
			  		+ "FROM (SELECT HID, Year "
					+ "		 FROM elementsWindows "
					+ "		 GROUP BY HID, Year "
					+ "		 HAVING COUNT(*)=2 AND SUM(AnnualCons)<0)");


		// Register the temporary table housesWithAtLeastOneDecrease 
		housesWithAtLeastOneDecreaseDF.createOrReplaceTempView("housesWithAtLeastOneDecrease");

		// Select the cities with at most two houses with at least one annual decrease 
		// removing from all cities those with a number of houses with at least one 
		// annual decrease greater than two. This approach also selects the cities with 
		// 0 houses with at least one annual decrease.
		Dataset<Row> res2DF = 
				ss.sql("SELECT DISTINCT City "
			  		+ "FROM Houses "
					+ "WHERE City NOT IN (SELECT City "
					+ "		 			  FROM Houses, housesWithAtLeastOneDecrease "
					+ "		 			  WHERE Houses.HID=housesWithAtLeastOneDecrease.HID "
					+ " 				  GROUP BY City "
					+ "					  HAVING COUNT(*)>2)");

		res2DF.write()
			.format("csv")
			.option("header", false)
			.save(outputPath2);
			

		// Close the Spark session
		ss.stop();
	}
}
