package it.polito.bigdata.spark.example;

import org.apache.spark.api.java.*;
import org.apache.spark.SparkConf;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import scala.Tuple2;
import scala.Tuple3;

import java.util.ArrayList;

public class SparkDriver {
	
	public static void main(String[] args) {

		// The following two lines are used to switch off some verbose log messages
		Logger.getLogger("org").setLevel(Level.OFF);
		Logger.getLogger("akka").setLevel(Level.OFF);


		String housePath, consumptionPath;
		String outputPath1, outputPath2;

		housePath = args[0];
		consumptionPath = args[1];
		outputPath1 = args[2];
		outputPath2 = args[3];

	
		SparkConf conf=new SparkConf().setAppName("Exam20220215 Spark")
		.setMaster("local");
		
		// Use the following command to create the SparkConf object if you want to run
		// your application inside Eclipse.
		// Remember to remove .setMaster("local") before running your application on the cluster
		// SparkConf conf=new SparkConf().setAppName("Spark Lab #5").setMaster("local");
		
		JavaSparkContext sc = new JavaSparkContext(conf);

		// HID,City,Country,YearBuilt
		JavaRDD<String> houseRDD = sc.textFile(housePath);
		// HID,Month,M3
		// Month in YYYY/MM format
		JavaRDD<String> consumptionRDD = sc.textFile(consumptionPath);

		// Part 1
		// filter water consumptions only for years 2022 and 2021
		JavaRDD<String> filteredCons = consumptionRDD
				.filter(l -> {
					String[] fields = l.split(",");
					String month = fields[1];

					return month.startsWith("2021") || month.startsWith("2022");
				});

		// compute the trimesters for each record
		// to do that, we use division between integers, with months starting from 0 (january = 0, february = 1 etc)
		// (month-1) / 3 = trimesterID associated to the record
		// example: january / 3 = 0, february / 3 = 0, march / 3 = 0, april / 3 = 1 etc.
		// We compute then an RDD with
		// key = HID, trimesterID, year
		// value = water consumption
		// and then sum all the values to compute the water consumed in that trimester
		JavaPairRDD<Tuple3<String, Integer, Integer>, Double> waterConsPerTrim =
				filteredCons
				.mapToPair(l -> {
					String[] fields = l.split(",");
					String hid = fields[0];
					String yearMonth = fields[1];
					Double consumption = Double.parseDouble(fields[2]);

					String[] f = yearMonth.split("/");
					Integer year = Integer.parseInt(f[0]);
					int trimester = (Integer.parseInt(f[1])-1) / 3;
					return new Tuple2<>(new Tuple3<>(hid, trimester, year), consumption);
				})
				.reduceByKey((v1, v2) -> v1 + v2);

		// perform a map to pair, obtaining
		// key = HID, trimester
		// value = (m3 in 2021, m3 in 2022)
		// where one of the two entries is 0
		// then, use a reduceByKey to sum the entries for each year (to obtain a single entry per house and per trimester)
		JavaPairRDD<Tuple2<String, Integer>, Tuple2<Double, Double>> houseConsPerTrim =
				waterConsPerTrim
				.mapToPair(t -> {
					String hid = t._1()._1();
					int trim = t._1()._2();
					int year = t._1()._3();
					double m3 = t._2();
					if (year == 2021)
						return new Tuple2<>(new Tuple2<>(hid, trim), new Tuple2<>(m3, 0.));
					else
						return new Tuple2<>(new Tuple2<>(hid, trim), new Tuple2<>(0., m3));
				})
				.reduceByKey((v1, v2) -> new Tuple2<Double, Double>(v1._1() + v2._1(), v1._2() + v2._2()));

		// convert the value part into
		// 0 if no increasing consumption
		// 1 if increasing consumption
		// and map the pairs with key = HID
		// and count for each house the number of increasing consumption associated to the same trimester
		// and filter with count >= 3
		JavaPairRDD<String, Integer> housesWithIncreasingCons =
				houseConsPerTrim
				.mapToPair(t -> new Tuple2<String, Integer>(t._1()._1(), t._2()._2() > t._2()._1() ? 1 : 0))
				.filter(p -> p._2()==1)
				.reduceByKey((v1, v2) -> v1 + v2)
				.filter(t -> t._2() >= 3);

		// prepare a pairRDD with
		// key = hid
		// value = city
		JavaPairRDD<String, String> houseCity = houseRDD.mapToPair(l -> {
			String[] fields = l.split(",");
			String hid = fields[0];
			String city = fields[1];
			return new Tuple2<>(hid, city);
		});

		// join the two RDDs to compute result1
		JavaPairRDD<String, String> res1 = houseCity
				.join(housesWithIncreasingCons)
				.mapToPair(t -> new Tuple2<>(t._1(), t._2()._1()));

		res1.saveAsTextFile(outputPath1);

		
		// Part 2
		// compute for each year and houseID, the water consumption
		// key = houseID, year
		// value = water consumption
		JavaPairRDD<Tuple2<String, Integer>, Double> waterConsPerYearAndHouse =
				consumptionRDD
				.mapToPair(l -> {
					String[] fields = l.split(",");
					String hid = fields[0];
					String yearMonth = fields[1];
					Double consumption = Double.parseDouble(fields[2]);
					String[] f = yearMonth.split("/");
					return new Tuple2<>(new Tuple2<>(hid, Integer.parseInt(f[0])), consumption);
				})
				.reduceByKey((v1, v2) -> v1 + v2);

		// compute an RDD with
		// key = HID
		// value = (year, m3)
		// and use a groupByKey to obtain a list, for each house, of all the 40 years of data
		// and sort such list based on year
		// Within the same lambda function, we verify whether or not there is at least one consumption decrease.
		// If true, return +1, else 0
		JavaPairRDD<String, Integer> houseConsumOverYears =
				waterConsPerYearAndHouse
				.mapToPair(t -> new Tuple2<>(t._1()._1(), new Tuple2<>(t._1()._2(), t._2())))
				.groupByKey()
				.mapValues(v -> {
					ArrayList<Tuple2<Integer, Double>> arr = new ArrayList<>();
					for(Tuple2<Integer, Double> element : v)
						arr.add(element);
					arr.sort((v1, v2) -> v1._1().compareTo(v2._1()));
					Double prevCons = arr.get(0)._2();
					for(Tuple2<Integer, Double> t : arr) {
						if(t._2() < prevCons)
							return 1;
					}
					return 0;
				});

		// Join the previously computed RDD with houseCityRDD to get information related to the city
		// and obtain the resulting RDD with
		// key = city
		// value = +1
		// and count the number of houses for each city for which an annual consumption decrease was recorded
		JavaPairRDD<String, Integer> cityHouseConsDecr =
				houseConsumOverYears
				.join(houseCity)
				.mapToPair(t -> new Tuple2<>(t._2()._2(), t._2()._1()))
				.reduceByKey((v1, v2) -> v1 + v2);

		// filter and keep only the cities with count <= 2
		JavaRDD<String> res2 =
				cityHouseConsDecr
				.filter(t -> t._2() <= 2)
				.keys();

		res2.saveAsTextFile(outputPath2);

		sc.close();
	}
}
